#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "game.h"
#include "config.h"
#include "log.h"
#include "world.h"
#include "voxels.h"
#include "generate.h"
#include "render.h"
#include "physics.h"
#include "player.h"

/* Keybindings for choosing tool. */
static unsigned short keybindings[36];

/* Clear tool keybinding. */
static void keys_clear() {
	int i;
	for (i = 0; i < sizeof(keybindings)/sizeof(keybindings[0]); i ++) {
		if (i+1 < LOADEDVOXELS) {
			keybindings[i] = i+1;
		} else {
			keybindings[i] = 1;
		}
	}
}

/* Load bindings, return true on success. */
static char keys_load() {
	errno = 0;
	FILE *fd = fopen(bind_path, "rb");
	if (fd == NULL) {
		LOG_PERROR("Couldn't open "BINDFILE);
		return 0;
	}

	int index;
	for (index = 0; index < sizeof(keybindings)/sizeof(keybindings[0]); index ++) {
		keybindings[index] = ((unsigned short)fgetc(fd)<<8) | ((unsigned short)fgetc(fd));
		if (errno) {
			LOG_PERROR("Error while reading "BINDFILE);
			fclose(fd);
			return 0;
		}
		if (keybindings[index] >= VOXELTYPES) {
			LOG_ERROR(BINDFILE" damaged, invalid voxel");
			fclose(fd);
			return 0;
		}
	}

	LOG_INFO("Read "BINDFILE);

	fclose(fd);
	return 1;
}

/* Save bindings. */
int keys_save() {
	errno = 0;
	FILE *fd = fopen(bind_path, "wb");
	if (fd == NULL) {
		LOG_PERROR("Couldn't create "BINDFILE);
		return 0;
	}

	int index;
	for (index = 0; index < sizeof(keybindings)/sizeof(keybindings[0]); index ++) {
		fputc(keybindings[index]>>16, fd); fputc(keybindings[index]&0xff, fd);
		if (errno) {
			LOG_PERROR("Error while reading "BINDFILE);
			fclose(fd);
			return 0;
		}
	}

	LOG_INFO("Wrote "BINDFILE);

	fclose(fd);
	return 1;
}

void game_init() {
	resource_init();
	render_init();
	if (!keys_load()) {
		keys_clear();
	}
	world_init();
	player_init();
	physics_init();
	voxels_init();
}

void game_parseHelp(const int argc, char *argv[]) {
	int i;
	for (i = 1; i < argc; i ++) {
		if (strcmp("-h", argv[i]) == 0 || strcmp("--help", argv[i]) == 0) {
			argv[i][0] = '\0';

			// NOTE: LOG_INFO used instead of fprintf because stderr seems to vanish on Windows
			LOG_INFO("Usage: build-minimal [OPTION]...Simple Minecraft Classic style game.");
			LOG_INFO("    -s, --sky      set sky colour {day, twilight, night, cave, desert}");
			LOG_INFO("    -c, --sea      set sea colour {ocean, cloud, grass, cave, sand}");
			LOG_INFO("    -g, --generate generate world {flat, island, flight, cave, desert, aether, empty} overrides sky and sea");
			LOG_INFO("    -h, --help     display this help and exit");
			LOG_INFO("    -v, --version  output version information and exit");
			exit(EXIT_SUCCESS);
		} else if (strcmp("-v", argv[i]) == 0 || strcmp("--version", argv[i]) == 0) {
			argv[i][0] = '\0';

			LOG_INFO("build-minimal "VERSION"\r\nCopyright(C) 2014-2015 Fudge the Sphinx\r\n");
			exit(EXIT_SUCCESS);
		}
	}
}

char game_parseGenerator(const int argc, char *argv[]) {
	int i;
	for (i = 1; i < argc; i ++) {
		if (strcmp("-g", argv[i]) == 0 || strcmp("--generate", argv[i]) == 0) {
			argv[i][0] = '\0';

			i ++;
			if (i >= argc) {
				LOG_ERROR("No landscape type specified");
				exit(EXIT_FAILURE);
			}

			if (strcmp("flat", argv[i]) == 0) {
				generate_flat(time(NULL));
			} else if (strcmp("island", argv[i]) == 0) {
				generate_island(time(NULL));
			} else if (strcmp("flight", argv[i]) == 0) {
				generate_flight(time(NULL));
			} else if (strcmp("cave", argv[i]) == 0) {
				generate_cave(time(NULL));
			} else if (strcmp("empty", argv[i]) == 0) {
				world_set(WORLDSIZE/2, WORLDSIZE/2, WORLDSIZE/2, VOXEL_DIRT);
				seaType = VOXEL_WATER;
			} else if (strcmp("desert", argv[i]) == 0) {
				generate_desert(time(NULL));
			} else if (strcmp("aether", argv[i]) == 0) {
				generate_aether(time(NULL));
			} else {
				LOG_ERROR("Invalid landscape type");
				exit(EXIT_FAILURE);
			}
			argv[i][0] = '\0';

			return 1;
		}
	}
	return 0;
}

void game_parseColours(const int argc, char *argv[]) {
	int i;
	for (i = 1; i < argc; i ++) {
		if (strcmp("-s", argv[i]) == 0 || strcmp("--sky", argv[i]) == 0) {
			argv[i][0] = '\0';

			i ++;
			if (i >= argc) {
				LOG_ERROR("No sky colour specified");
				exit(EXIT_FAILURE);
			}

			if (strcmp("day", argv[i]) == 0) {
				skyHorizon = 0x9BCCE9;
				skyAtmosphere = 0x6AA8D3;
				skyZenith = 0x366AA3;
				sunColour = 0xFFFFFF;
			} else if (strcmp("twilight", argv[i]) == 0) {
				skyHorizon = 0xE27C40;
				skyAtmosphere = 0x60789F;
				skyZenith = 0x1D283C;
				sunColour = 0xB4AE9D;
			} else if (strcmp("night", argv[i]) == 0) {
				skyHorizon = 0x1B374D;
				skyAtmosphere = 0x051A2F;
				skyZenith = 0x000612;
				sunColour = 0x434548;
			} else if (strcmp("cave", argv[i]) == 0) {
				skyHorizon = 0x565A57;
				skyAtmosphere = 0x434443;
				skyZenith = 0x161616;
				sunColour = 0x808080;
			} else if (strcmp("desert", argv[i]) == 0) {
				skyHorizon = 0xDCD7A7;
				skyAtmosphere = 0x9CD0DA;
				skyZenith = 0xF6F4E6;
				sunColour = 0xFFFFFF;
			} else if (strcmp("black", argv[i]) == 0) {
				skyHorizon = 0x000000;
				skyAtmosphere = 0x000000;
				skyZenith = 0x000000;
				sunColour = 0x000000;
			} else {
				LOG_ERROR("Invalid sky colour");
				exit(EXIT_FAILURE);
			}
			argv[i][0] = '\0';
		} else if (strcmp("-c", argv[i]) == 0 || strcmp("--sea", argv[i]) == 0) {
			argv[i][0] = '\0';

			i ++;
			if (i >= argc) {
				LOG_ERROR("No sea colour specified");
				exit(EXIT_FAILURE);
			}

			if (strcmp("ocean", argv[i]) == 0) {
				seaType = VOXEL_WATER;
			} else if (strcmp("cloud", argv[i]) == 0) {
				seaType = VOXEL_CUMULUS;
			} else if (strcmp("grass", argv[i]) == 0) {
				seaType = VOXEL_GRASS;
			} else if (strcmp("cave", argv[i]) == 0) {
				seaType = VOXEL_STONE;
			} else if (strcmp("sand", argv[i]) == 0) {
				seaType = VOXEL_SAND;
			} else {
				LOG_ERROR("Invalid sea colour");
				exit(EXIT_FAILURE);
			}
			argv[i][0] = '\0';
		}
	}
}

void game_parseJunk(const int argc, char *argv[]) {
	int i;
	for (i = 1; i < argc; i ++) {
		if (argv[i][0] != '\0') {
			LOG_ERROR("Unrecognised option:");
			LOG_ERROR(argv[i]);
			exit(EXIT_FAILURE);
		}
	}
}

void game_loop() {
	// Camera rotation speed for keyboard control (pan/elevation)
	float cameraDP = 0;
	float cameraDE = 0;

	// Time of last frame
	Uint32 lastframe = 0;

	// Temporary values
	int delta;
	SDL_Event e;
	const Uint8 *key = SDL_GetKeyboardState(NULL);

	// Clear state
	world_dirty();
	render_resize();
	render_updatesky(skyHorizon, skyAtmosphere, skyZenith);
	player_clear();
	render_update();

	while (1) {
		// Handle input
		while (SDL_PollEvent(&e)) {
			switch (e.type) {
				case SDL_QUIT:
					return;

				case SDL_KEYDOWN:
					if (e.key.keysym.sym == SDLK_ESCAPE) {
						if (e.key.keysym.mod&KMOD_SHIFT) {
							// Shift+Esc - toggle mouse grab
							if (SDL_GetRelativeMouseMode()) {
								SDL_SetRelativeMouseMode(SDL_FALSE);
							} else {
								SDL_SetRelativeMouseMode(SDL_TRUE);
							}
						} else {
							// Ctrl+Esc - exit without saving
							// Esc - exit and save
							if ((e.key.keysym.mod & KMOD_CTRL) == 0) {
								world_save(world_path);
							}
							return;
						}
					} else if (e.key.keysym.sym == SDLK_EQUALS) {
						if (e.key.keysym.mod&KMOD_CTRL) {
							player_switch(10);
						} else {
							player_switch(1);
						}
					} else if (e.key.keysym.sym == SDLK_MINUS) {
						if (e.key.keysym.mod&KMOD_CTRL) {
							player_switch(-10);
						} else {
							player_switch(-1);
						}
					} else if (e.key.keysym.sym == SDLK_z) {
						if (e.key.keysym.mod&KMOD_CTRL) {
							player_create();
						} else if (e.key.keysym.mod&KMOD_SHIFT) {
							player_select();
						} else {
							player_destroy();
						}
					} else if (e.key.keysym.sym == SDLK_x) {
						player_create();
					} else if (e.key.keysym.sym == SDLK_c) {
						if (e.key.keysym.mod&KMOD_CTRL) {
							player_select();
						} else {
							player_activate();
						}
					} else if (e.key.keysym.sym == SDLK_TAB) {
						player_fly();
					} else {
						// Select block by keybinding or update keybinding
						int bindindex = e.key.keysym.sym;
						if (bindindex >= SDLK_0 && bindindex <= SDLK_9) {
							bindindex = bindindex-SDLK_0;
						} else if (bindindex >= SDLK_a && bindindex <= SDLK_z &&
								bindindex != SDLK_q &&
								bindindex != SDLK_e &&
								bindindex != SDLK_w &&
								bindindex != SDLK_a &&
								bindindex != SDLK_s &&
								bindindex != SDLK_d &&
								bindindex != SDLK_z &&
								bindindex != SDLK_x &&
								bindindex != SDLK_c) {
							bindindex = bindindex-SDLK_a+10;
						} else {
							bindindex = -1;
						}

						if (bindindex != -1) {
							if ((e.key.keysym.mod&KMOD_CTRL) != 0) {
								keybindings[bindindex] = player_tool;
							} else {
								player_tool = keybindings[bindindex];
							}
						}
					}
					break;

				case SDL_MOUSEMOTION:
					if (SDL_GetRelativeMouseMode()) {
						player_look((float)e.motion.xrel * MOUSESPEED, (float)e.motion.yrel * MOUSESPEED);
					}
					break;

				case SDL_MOUSEBUTTONDOWN:
					if (SDL_GetRelativeMouseMode()) {
						if (e.button.button == SDL_BUTTON_LEFT) {
							player_destroy();
						} else if (e.button.button == SDL_BUTTON_MIDDLE) {
							player_select();
						} else if (e.button.button == SDL_BUTTON_RIGHT) {
							player_create();
						}
					} else {
						SDL_SetRelativeMouseMode(SDL_TRUE);
					}
					break;

				case SDL_MOUSEWHEEL:
					if (e.wheel.y < 0 || e.wheel.x < 0) {
						player_switch(-1);
					} else if (e.wheel.y > 0 || e.wheel.x > 0) {
						player_switch(1);
					}
					break;

				case SDL_WINDOWEVENT:
					if (e.window.event == SDL_WINDOWEVENT_RESIZED) {
						render_resize();
					}
					break;
			}
		}

		// Input
		if (key[SDL_SCANCODE_LEFT]) {
			cameraDP -= 0.7;
		} else if (key[SDL_SCANCODE_RIGHT]) {
			cameraDP += 0.7;
		}
		if (key[SDL_SCANCODE_UP]) {
			cameraDE -= 0.7;
		} else if (key[SDL_SCANCODE_DOWN]) {
			cameraDE += 0.7;
		}
		player_look(cameraDP, cameraDE);
		cameraDP *= 0.8;
		cameraDE *= 0.8;
		player_move(key[SDL_SCANCODE_D] - key[SDL_SCANCODE_A],
			(key[SDL_SCANCODE_SPACE] | key[SDL_SCANCODE_E]) - key[SDL_SCANCODE_Q],
			key[SDL_SCANCODE_W] - key[SDL_SCANCODE_S]);
		player_update();

		// Physics
		physics_update();

		// Lighting
		world_relight();

		// Render
		render_normal(
			cameraX, cameraY, cameraZ, cameraP, cameraT,
			selected,
			selectX, selectY, selectZ,
			faceX, faceY, faceZ,
			player_tool);

		// Limit framerate
		delta = SDL_GetTicks()-lastframe;
		if (delta < 1000/30) {
			SDL_Delay(1000/30 - delta);
		}
		lastframe = SDL_GetTicks();
	}
}

void game_cleanup() {
	keys_save();
	resource_quit();
}

