#include <stdlib.h>
#include <SDL2/SDL.h>
#include "physics.h"
#include "config.h"
#include "world.h"
#include "render.h"
#include "voxels.h"

void (*physics_updatefn[VOXELTYPES])(unsigned short type, int x, int y, int z);
void (*physics_changefn[VOXELTYPES])(unsigned short type, int x, int y, int z, char dx, char dy, char dz);
char (*physics_destroyfn[VOXELTYPES])(unsigned short type, int x, int y, int z, char dx, char dy, char dz);
void (*physics_activatefn[VOXELTYPES])(unsigned short type, int x, int y, int z, char dx, char dy, char dz);
void (*physics_tool_destroyfn[VOXELTYPES])(unsigned short selected, int x, int y, int z, char dx, char dy, char dz);
void (*physics_tool_selectfn[VOXELTYPES])(unsigned short selected, int x, int y, int z, char dx, char dy, char dz);
void (*physics_tool_createfn[VOXELTYPES])(unsigned short selected, int x, int y, int z, char dx, char dy, char dz);

void physics_init() {
	size_t index;
	for (index = 0; index < VOXELTYPES; index ++) {
		physics_updatefn[index] = NULL;
		physics_changefn[index] = NULL;
		physics_destroyfn[index] = NULL;
		physics_activatefn[index] = NULL;
		physics_tool_destroyfn[index] = NULL;
		physics_tool_selectfn[index] = NULL;
		physics_tool_createfn[index] = NULL;
	}
}

void physics_update() {
	int x, y, z, type;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (y = 0; y < WORLDSIZE; y ++) {
			for (z = 0; z < WORLDSIZE; z ++) {
				type = world_get(x, y, z);
				if (physics_updatefn[type] != NULL) {
					physics_updatefn[type](type, x, y, z);
				}
			}
		}
	}
}

void physics_changed(const int x, const int y, const int z) {
#ifndef DISABLEPHYSICS
	unsigned short type;
	if (x > 0) {
		type = world_get(x-1, y, z);
		if (physics_changefn[type] != NULL) {
			physics_changefn[type](type, x-1, y, z, 1, 0, 0);
		}
	}
	if (x < WORLDSIZE-1) {
		type = world_get(x+1, y, z);
		if (physics_changefn[type] != NULL) {
			physics_changefn[type](type, x+1, y, z, -1, 0, 0);
		}
	}
	if (y > 0) {
		type = world_get(x, y-1, z);
		if (physics_changefn[type] != NULL) {
			physics_changefn[type](type, x, y-1, z, 0, 1, 0);
		}
	}
	if (y < WORLDSIZE-1) {
		type = world_get(x, y+1, z);
		if (physics_changefn[type] != NULL) {
			physics_changefn[type](type, x, y+1, z, 0, -1, 0);
		}
	}
	if (z > 0) {
		type = world_get(x, y, z-1);
		if (physics_changefn[type] != NULL) {
			physics_changefn[type](type, x, y, z-1, 0, 0, 1);
		}
	}
	if (z < WORLDSIZE-1) {
		type = world_get(x, y, z+1);
		if (physics_changefn[type] != NULL) {
			physics_changefn[type](type, x, y, z+1, 0, 0, -1);
		}
	}
#endif
}

/* Get voxel. */
static unsigned short get(const int x, const int y, const int z) {
	// Sea
	if (y < 0) {
		return seaType;
	}

	// Air
	if (x < 0 || x >= WORLDSIZE || y >= WORLDSIZE || z < 0 || z >= WORLDSIZE) {
		return VOXEL_AIR;
	}

	return world_get(x, y, z);
}

/* Change a voxel without updating any graphics or lighting.
WARNING: use carefully */
static void set_invisible(const int x, const int y, const int z, const unsigned short type) {
	if (x < 0 || x >= WORLDSIZE || y < 0 || y >= WORLDSIZE || z < 0 || z >= WORLDSIZE) {
		return;
	}
	world_set(x, y, z, type);
	physics_changed(x, y, z);
}

/* Change a voxel. */
static void set(const int x, const int y, const int z, const unsigned short type) {
	if (x < 0 || x >= WORLDSIZE || y < 0 || y >= WORLDSIZE || z < 0 || z >= WORLDSIZE) {
		return;
	}
	world_set(x, y, z, type);
	physics_changed(x, y, z);
	world_dirty();
	render_update();
}

/* Replace a group of the same type of voxel. */
static void fill(const int x, const int y, const int z, const unsigned short current, const unsigned short new) {
	if (x < 0 || x >= WORLDSIZE || y < 0 || y >= WORLDSIZE || z < 0 || z >= WORLDSIZE) {
		return;
	}
	if (world_get(x, y, z) != current) {
		return;
	}
	set(x, y, z, new);
	fill(x-1, y, z, current, new);
	fill(x+1, y, z, current, new);
	fill(x, y-1, z, current, new);
	fill(x, y+1, z, current, new);
	fill(x, y, z-1, current, new);
	fill(x, y, z+1, current, new);
}

void physics_tool_none(unsigned short selected, int x, int y, int z, char dx, char dy, char dz) {
}

void physics_door_open(unsigned short type, int x, int y, int z, char dx, char dy, char dz) {
	SDL_assert(type < LOADEDVOXELS-1);
	fill(x, y, z, type, type+1);
}

void physics_door_close(unsigned short type, int x, int y, int z, char dx, char dy, char dz) {
	SDL_assert(type > 0);
	fill(x, y, z, type, type-1);
}

/* Return true if block is lit by ambient light. */
static char lit(const int x, int y, const int z) {
	for (y ++; y < WORLDSIZE; y ++) {
		if (opaque[get(x, y, z)]) {
			return 0;
		}
	}
	return 1;
}

void physics_grass(unsigned short type, int x, int y, int z) {
	if (rand()%1000 > 0) {
		return;
	}

	if (!lit(x, y, z)) {
		set(x, y, z, VOXEL_DIRT);
		return;
	}

	// Spread
	if (get(x-1, y, z) == VOXEL_DIRT && lit(x-1, y, z)) {
		set(x-1, y, z, VOXEL_GRASS);
	}
	if (get(x+1, y, z) == VOXEL_DIRT && lit(x+1, y, z)) {
		set(x+1, y, z, VOXEL_GRASS);
	}
	if (get(x, y, z-1) == VOXEL_DIRT && lit(x, y, z-1)) {
		set(x, y, z-1, VOXEL_GRASS);
	}
	if (get(x, y, z+1) == VOXEL_DIRT && lit(x, y, z+1)) {
		set(x, y, z+1, VOXEL_GRASS);
	}

	if (rand()%10 == 0 && get(x, y+1, z) == VOXEL_AIR) {
		if (rand()%10 == 0) {
			// Plant
			set(x, y+1, z, VOXEL_STEM + rand()%4);
		} else {
			// Long grass
			set(x, y+1, z, VOXEL_LARGEGRASS);
		}
	}
}

void physics_wheat(unsigned short type, int x, int y, int z) {
	if (rand()%1000 > 0) {
		return;
	}

	// Check if lit
	int test;
	for (test = y + 1; test < WORLDSIZE; test ++) {
		if (opaque[get(x, test, z)]) {
			return;
		}
	}

	if (rand()%10 == 0) {
		if (type < VOXEL_WHEAT3) {
			set(x, y, z, type+1);
		}
	}
}

/* Add a value to a stack if it isn't in the stack already. */
static void append(int *stackx, int *stackz, char *stackdx, char *stackdz, size_t *length, const int x, const int z, const char dx, const char dz) {
	size_t index;
	for (index = 0; index < *length; index ++) {
		if (stackx[index] == x && stackz[index] == z) {
			// Voxel is already in the stack
			return;
		}
	}

	// Add voxel to stack
	stackx[*length] = x;
	stackz[*length] = z;
	stackdx[*length] = dx;
	stackdz[*length] = dz;
	(*length) ++;
}

/* Get direction of the closest slope. */
static char find_slope(const int x, const int y, const int z, char *dx, char *dz) {
	SDL_assert(y > 0);

	// Stack of voxels to test
	int stackx[WORLDSIZE*WORLDSIZE];
	int stackz[WORLDSIZE*WORLDSIZE];
	char stackdx[WORLDSIZE*WORLDSIZE];
	char stackdz[WORLDSIZE*WORLDSIZE];
	size_t length = 0;
	size_t next = 0;

	// Add first voxels
	if (x > 0 && get(x-1, y, z) == VOXEL_AIR) {
		stackx[length] = x-1;
		stackz[length] = z;
		stackdx[length] = -1;
		stackdz[length] = 0;
		length ++;
	}

	if (x < WORLDSIZE-1 && get(x+1, y, z) == VOXEL_AIR) {
		stackx[length] = x+1;
		stackz[length] = z;
		stackdx[length] = 1;
		stackdz[length] = 0;
		length ++;
	}

	if (z > 0 && get(x, y, z-1) == VOXEL_AIR) {
		stackx[length] = x;
		stackz[length] = z-1;
		stackdx[length] = 0;
		stackdz[length] = -1;
		length ++;
	}

	if (z < WORLDSIZE-1 && get(x, y, z+1) == VOXEL_AIR) {
		stackx[length] = x;
		stackz[length] = z+1;
		stackdx[length] = 0;
		stackdz[length] = 1;
		length ++;
	}

	// Find nearest slope
	int u, v;
	while (next != length) {
		u = stackx[next];
		v = stackz[next];

		// Check if there is a space below the voxel
		if (get(u, y-1, v) == VOXEL_AIR) {
			// Found the nearest slope, return data
			*dx = stackdx[next];
			*dz = stackdz[next];
			return 1;
		}

		// Add adjacent voxels to stack if they aren't already in the stack and they are empty
		if (u > 0 && get(u-1, y, v) == VOXEL_AIR) {
			append(stackx, stackz, stackdx, stackdz, &length, u-1, v, stackdx[next], stackdz[next]);
		}
		if (u < WORLDSIZE-1 && get(u+1, y, v) == VOXEL_AIR) {
			append(stackx, stackz, stackdx, stackdz, &length, u+1, v, stackdx[next], stackdz[next]);
		}
		if (v > 0 && get(u, y, v-1) == VOXEL_AIR) {
			append(stackx, stackz, stackdx, stackdz, &length, u, v-1, stackdx[next], stackdz[next]);
		}
		if (v < WORLDSIZE-1 && get(u, y, v+1) == VOXEL_AIR) {
			append(stackx, stackz, stackdx, stackdz, &length, u, v+1, stackdx[next], stackdz[next]);
		}
		next ++;
	}

	return 0;
}

/* Recursively look for liquid above the current layer. */
static char get_pressure_recurse(const unsigned short type, const int x, const int y, const int z, char tested[WORLDSIZE][WORLDSIZE]) {
	if (get(x, y, z) != type || tested[x][z]) {
		return 0;
	}
	if (get(x, y+1, z) == type) {
		return 1;
	}
	tested[x][z] = 1;
	return get_pressure_recurse(type, x-1, y, z, tested) ||
		get_pressure_recurse(type, x+1, y, z, tested) ||
		get_pressure_recurse(type, x, y, z-1, tested) ||
		get_pressure_recurse(type, x, y, z+1, tested);
}

/* Return true if there is liquid on top of this layer forcing it to flow. */
static char get_pressure(const unsigned short type, const int x, const int y, const int z) {
	// Matrix of booleans to record which voxels have been checked
	char tested[WORLDSIZE][WORLDSIZE];
	int u, v;
	for (u = 0; u < WORLDSIZE; u ++) {
		for (v = 0; v < WORLDSIZE; v ++) {
			tested[u][v] = 0;
		}
	}

	// Check for water above this water
	return get_pressure_recurse(type, x, y, z, tested);
}

/* Flow. */
static void physics_liquid(const unsigned short type, const int x, const int y, const int z) {
	// Fall out of the bottom of the world
	if (y == 0) {
		set(x, y, z, VOXEL_AIR);
		return;
	}

	// Try to fall
	if (get(x, y-1, z) == VOXEL_AIR) {
		set(x, y-1, z, type);
		set(x, y, z, VOXEL_AIR);
		return;
	}

	// Find nearest slope and move in that direction
	char dx, dz;
	if (find_slope(x, y, z, &dx, &dz)) {
		SDL_assert(get(x+dx, y, z+dz) == VOXEL_AIR);
		set(x+dx, y, z+dz, type);
		set(x, y, z, VOXEL_AIR);
		return;
	}

	// Move sideways if there is water above this layer of water
	if (y < WORLDSIZE-1 && get_pressure(type, x, y, z)) {
		// Get random direction to test first
		dx = (rand()%2)*2 - 1;
		dz = (rand()%2)*2 - 1;
		if (rand()%2 == 0) {
			dx = 0;
		} else {
			dz = 0;
		}

		// Look for a direction to move
		char count;
		for (count = 0; count < 4; count ++) {
			if (get(x+dx, y, z+dz) == VOXEL_AIR) {
				// Move to new location
				set(x+dx, y, z+dz, type);
				set(x, y, z, VOXEL_AIR);
				return;
			}
		}
	}
}

void physics_water(unsigned short type, int x, int y, int z) {
	if (rand()%3 > 0) {
		return;
	}

	physics_liquid(type, x, y, z);
}

void physics_lava(unsigned short type, int x, int y, int z) {
	if (rand()%10 > 0) {
		return;
	}

	// Fuse if cooled by water
	if (get(x-1, y, z) == VOXEL_WATER ||
			get(x+1, y, z) == VOXEL_WATER ||
			get(x, y-1, z) == VOXEL_WATER ||
			get(x, y+1, z) == VOXEL_WATER ||
			get(x, y, z-1) == VOXEL_WATER ||
			get(x, y, z+1) == VOXEL_WATER) {
		set(x, y, z, VOXEL_STONE);
		return;
	}

	// Move
	physics_liquid(type, x, y, z);
}

void physics_sand(unsigned short type, int x, int y, int z) {
	if (rand()%2 > 0) {
		return;
	}

	// Fall
	unsigned short on = get(x, y-1, z);
	if (on == VOXEL_AIR || on == VOXEL_WATER) {
		set(x, y-1, z, type);
		set(x, y, z, on);
	}
}
