#include <stdio.h>
#include <math.h>
#include <SDL2/SDL_image.h>
#include "perspectiveGL.h"
#include "config.h"
#include "log.h"
#include "render.h"

// Number of textures packed into atlas on each axis
#define ATLASSIZE 64
// Size of each texture in the atlas in uv coordinates
#define TEXSIZE 1.0/ATLASSIZE

/* VBO names. */
static GLuint coords, sky;
/* Buffer to build VBOs. */
#ifndef DISABLELIGHTING
// Number of voxels * six faces per voxel * four verticies per face * (three coordinates + two texture coordinates)
static GLfloat buffer[WORLDSIZE*WORLDSIZE*WORLDSIZE*6*4*5];
#else
// Number of voxels * six faces per voxel * four verticies per face * (three coordinates + two texture coordinates + three colours for lights)
static GLfloat buffer[WORLDSIZE*WORLDSIZE*WORLDSIZE*6*4*8];
#endif
/* Number of quads in VBO. */
static GLsizei quads;

/* Set when the VBO needs to be rebuilt. */
static unsigned char rebuild = 0;

unsigned int topMap[VOXELTYPES];
unsigned int bottomMap[VOXELTYPES];
unsigned int leftMap[VOXELTYPES];
unsigned int rightMap[VOXELTYPES];
unsigned int frontMap[VOXELTYPES];
unsigned int backMap[VOXELTYPES];
unsigned char specialMap[VOXELTYPES];

/* Viewport size. */
static int width, height;

void render_updatesky(const unsigned long horizoncolour, const unsigned long atmospherecolour, const unsigned long zenithcolour) {
	// Calculate colours
	GLfloat haze[] = {(horizoncolour>>16)/255.0, ((horizoncolour>>8)&0xff)/255.0, (horizoncolour&0xff)/255.0, 1};
	GLfloat horizon[] = {(atmospherecolour>>16)/255.0, ((atmospherecolour>>8)&0xff)/255.0, (atmospherecolour&0xff)/255.0};
	GLfloat above[] = {(zenithcolour>>16)/255.0, ((zenithcolour>>8)&0xff)/255.0, (zenithcolour&0xff)/255.0};

	#ifndef DISABLEFOG
	// Setup fog colour
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogfv(GL_FOG_COLOR, haze);
	glFogf(GL_FOG_START, 30);
	glFogf(GL_FOG_END, 70);
	#endif

	// Calculate mesh
	GLfloat skydata[120] = {
		// Top
		-50, 0.8, -0.4, above[0], above[1], above[2],
		50, 0.8, -0.4, above[0], above[1], above[2],
		50, 0.8, 1, above[0], above[1], above[2],
		-50, 0.8, 1, above[0], above[1], above[2],

		// Bottom
		-50, -0.8, -0.4, haze[0], haze[1], haze[2],
		-50, -0.8, 1, haze[0], haze[1], haze[2],
		50, -0.8, 1, haze[0], haze[1], haze[2],
		50, -0.8, -0.4, haze[0], haze[1], haze[2],

		// Back
		50, -0.8, -0.4, haze[0], haze[1], haze[2],
		50, -0.15, -0.4, haze[0], haze[1], haze[2],
		-50, -0.15, -0.4, haze[0], haze[1], haze[2],
		-50, -0.8, -0.4, haze[0], haze[1], haze[2],

		50, -0.15, -0.4, haze[0], haze[1], haze[2],
		50, 0.15, -0.4, horizon[0], horizon[1], horizon[2],
		-50, 0.15, -0.4, horizon[0], horizon[1], horizon[2],
		-50, -0.15, -0.4, haze[0], haze[1], haze[2],

		50, 0.15, -0.4, horizon[0], horizon[1], horizon[2],
		50, 0.8, -0.4, above[0], above[1], above[2],
		-50, 0.8, -0.4, above[0], above[1], above[2],
		-50, 0.15, -0.4, horizon[0], horizon[1], horizon[2]};

	// Load mesh
	glBindBuffer(GL_ARRAY_BUFFER, sky);
	glBufferData(GL_ARRAY_BUFFER, 120*sizeof(GLfloat), &skydata[0], GL_STATIC_DRAW);

	SDL_assert(glGetError() == GL_NO_ERROR);
}

void render_init() {
	// Set rendering state
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glClearColor(0, 0, 0, 1);
	glClearDepth(1);
	glDepthFunc(GL_LEQUAL);
	glAlphaFunc(GL_GREATER, 0.5);

	// Allocate VBOs
	glGenBuffers(1, &coords);
	glGenBuffers(1, &sky);

	// Clear texture map
	int i;
	for (i = 0; i < VOXELTYPES; i ++) {
		topMap[i] = 0;
		bottomMap[i] = 0;
		leftMap[i] = 0;
		rightMap[i] = 0;
		frontMap[i] = 0;
		backMap[i] = 0;
		specialMap[i] = 0;
	}
}

void render_resize() {
	SDL_GetWindowSize(window, &width, &height);
	SDL_assert_release(width > 0 && height > 0);

	GLfloat ratio = (GLfloat)width / (GLfloat)height;

	// Setup viewport
	glViewport(0, 0, (GLsizei)width, (GLsizei)height);

	// Setup projection
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	perspectiveGL(FOV, ratio, 0.1, 100);

	// Empty view
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	SDL_assert(glGetError() == GL_NO_ERROR);
}

void render_update() {
	rebuild = 1;
}

#ifndef DISABLELIGHTING
/* Sets *r, *g, *b to the lighting for the specified voxel. */
static void getlight(int x, int y, int z, GLfloat *r, GLfloat *g, GLfloat *b) {
	if (x < 0 || x >= WORLDSIZE || y < 0 || y >= WORLDSIZE || z < 0 || z >= WORLDSIZE) {
		*r = (GLfloat)(sunColour>>16)/0xff;
		*g = (GLfloat)((sunColour>>8)&0xff)/0xff;
		*b = (GLfloat)(sunColour&0xff)/0xff;
	} else {
		int index = (x*WORLDSIZE*WORLDSIZE+y*WORLDSIZE+z)*3;
		*r = (GLfloat)light[index]/0xff;
		*g = (GLfloat)light[index+1]/0xff;
		*b = (GLfloat)light[index+2]/0xff;
	}
}

/* Get the smooth illumination for one vertex of a face of a voxel.
x, y, z position of the voxel in front of the face
ux, uy, uz vector in direction of one adjacent face
vx, vy, vz vector in direction of other adjacent face */
static void calculate_light(
		int x, int y, int z,
		char ux, char uy, char uz,
		char vx, char vy, char vz,
		GLfloat *red, GLfloat *green, GLfloat *blue) {
	GLfloat addr, addg, addb;
	getlight(x, y, z, &addr, &addg, &addb);
	*red = addr; *green = addg; *blue = addb;

	getlight(x+ux, y+uy, z+uz, &addr, &addg, &addb);
	*red += addr; *green += addg; *blue += addb;

	getlight(x+vx, y+vy, z+vz, &addr, &addg, &addb);
	*red += addr; *green += addg; *blue += addb;

	getlight(x+ux+vx, y+uy+vy, z+uz+vz, &addr, &addg, &addb);
	*red += addr; *green += addg; *blue += addb;

	*red /= 4; *green /= 4; *blue /= 4;
}
#endif

/* Returns true if the specified voxel is solid. */
static unsigned char getsolid(int x, int y, int z) {
	if (x < 0 || x >= WORLDSIZE || y < 0 || y >= WORLDSIZE || z < 0 || z >= WORLDSIZE) {
		return 0;
	}
	return solid[world_get(x, y, z)];
}

/* Return offset of texture to apply connected texture. */
static int map(const char top, const char right, const char bottom, const char left,
		const char tr, const char br, const char bl, const char tl) {
	int edges = top+right+bottom+left;
	if (edges == 0) {
		return 0;
	} else if (edges == 1) {
		if (top) {
			return 1;
		} else if (right) {
			return 2;
		} else if (bottom) {
			return 3;
		} else {
			return 4;
		}
	} else if (edges == 2) {
		if (top && right) {
			return 5 + tr;
		} else if (right && bottom) {
			return 7 + br;
		} else if (bottom && left) {
			return 9 + bl;
		} else if (left && top) {
			return 11 + tl;
		} else if (top && bottom) {
			return 13;
		} else {
			return 14;
		}
	} else if (edges == 3) {
		if (left && top && right) {
			return 15 + tl + (tr<<1);
		} else if (top && right && bottom) {
			return 19 + tr + (br<<1);
		} else if (right && bottom && left) {
			return 23 + br + (bl<<1);
		} else {
			return 27 + bl + (tl<<1);
		}
	} else {
		return 31 + tl + (tr<<1) + (br<<2) + (bl<<3);
	}
	abort();
	return 0;
}

/* Return true if voxel at given location will connect to specified voxel type. */
static char willconnect(const int x, const int y, const int z, const unsigned short type) {
	if (x < 0 || x >= WORLDSIZE || y < 0 || y >= WORLDSIZE || z < 0 || z >= WORLDSIZE) {
		return 0;
	}
	return world_get(x, y, z) == type;
}

/* Add one side of a voxel to the buffer.
x, y, z position of voxel
ux, uy, uz rightward vector
vx, vy, vz upward vector
wx, wy, wz outward vector
index index of next position to store geometry
special is the type of rendering */
static void add_quad(const unsigned short type,
		const int x, const int y, const int z,
		const char ux, const char uy, const char uz,
		char vx, char vy, char vz,
		char wx, char wy, char wz,
		const unsigned char special,
		unsigned int texture,
		size_t *index) {
	#ifndef DISABLELIGHTING
	GLfloat red, green, blue;
	#endif

	// Check if texture empty
	if (texture == 0) {
		return;
	}
	texture --;

	// Check if hidden
	if (special != RENDER_X) {
		if (x+wx >= 0 && x+wx <= WORLDSIZE-1 &&
				y+wy >= 0 && y+wy <= WORLDSIZE-1 &&
				z+wz >= 0 && z+wz <= WORLDSIZE-1 &&
				(opaque[world_get((x+wx), (y+wy), (z+wz))] ||
					world_get((x+wx), (y+wy), (z+wz)) == type)) {
			return;
		}

		if (special == RENDER_JOIN) {
			// Use joined texture if face is surrounded by solid blocks
			if (getsolid(x-ux, y-uy, z-uz) &&
					getsolid(x+ux, y+uy, z+uz) &&
					getsolid(x-vx, y-vy, z-vz) &&
					getsolid(x+vx, y+vy, z+vz)) {
				texture ++;
			}
		} else if (special == RENDER_CONNECTED) {
			// Use connected texture
			texture += map(
				willconnect(x+vx, y+vy, z+vz, type), willconnect(x+ux, y+uy, z+uz, type), willconnect(x-vx, y-vy, z-vz, type), willconnect(x-ux, y-uy, z-uz, type),
				willconnect(x+ux+vx, y+uy+vy, z+uz+vz, type), willconnect(x+ux-vx, y+uy-vy, z+uz-vz, type), willconnect(x-ux-vx, y-uy-vy, z-uz-vz, type), willconnect(x-ux+vx, y-uy+vy, z-uz+vz, type));
		}
	}

	// Get coordinates of texture
	GLfloat texx = TEXSIZE*(texture%ATLASSIZE);
	GLfloat texy = TEXSIZE*(texture/ATLASSIZE);

	// Move first vertex to make orientation of quads consistent
	// sy is a float because the flat part of crosses are half way down
	// The other coordinates are always integer
	int sx = x;
	GLfloat sy = y;
	int sz = z;
	if (wx > 0 || ux < 0 || vx < 0) {
		sx ++;
	}
	if (wy > 0 || uy < 0 || vy < 0) {
		sy ++;
	}
	if (wz > 0 || uz < 0 || vz < 0) {
		sz ++;
	}

	// Append quad
	if (special == RENDER_X) {
		if (wy > 0) {
			wy = 0;
			sy -= 0.5;
		} else if (wy < 0) {
			wy = 0;
			sy += 0.1;
		}
		#ifndef DISABLELIGHTING
		getlight(x, y, z, &red, &green, &blue);
		#endif
		buffer[(*index)++] = sx; buffer[(*index)++] = sy; buffer[(*index)++] = sz; buffer[(*index)++] = texx; buffer[(*index)++] = texy+TEXSIZE;
		#ifndef DISABLELIGHTING
		buffer[(*index)++] = red; buffer[(*index)++] = green; buffer[(*index)++] = blue;
		#endif
		buffer[(*index)++] = sx+ux-wx; buffer[(*index)++] = sy+uy-wy; buffer[(*index)++] = sz+uz-wz; buffer[(*index)++] = texx+TEXSIZE; buffer[(*index)++] = texy+TEXSIZE;
		#ifndef DISABLELIGHTING
		buffer[(*index)++] = red; buffer[(*index)++] = green; buffer[(*index)++] = blue;
		#endif
		buffer[(*index)++] = sx+ux+vx-wx; buffer[(*index)++] = sy+uy+vy-wy; buffer[(*index)++] = sz+uz+vz-wz; buffer[(*index)++] = texx+TEXSIZE; buffer[(*index)++] = texy;
		#ifndef DISABLELIGHTING
		buffer[(*index)++] = red; buffer[(*index)++] = green; buffer[(*index)++] = blue;
		#endif
		buffer[(*index)++] = sx+vx; buffer[(*index)++] = sy+vy; buffer[(*index)++] = sz+vz; buffer[(*index)++] = texx; buffer[(*index)++] = texy;
		#ifndef DISABLELIGHTING
		buffer[(*index)++] = red; buffer[(*index)++] = green; buffer[(*index)++] = blue;
		#endif
	} else {
		buffer[(*index)++] = sx; buffer[(*index)++] = sy; buffer[(*index)++] = sz; buffer[(*index)++] = texx; buffer[(*index)++] = texy+TEXSIZE;
		#ifndef DISABLELIGHTING
		calculate_light(
			x+wx, y+wy, z+wz,
			-ux, -uy, -uz,
			-vx, -vy, -vz,
			&buffer[*index], &buffer[*index + 1], &buffer[*index + 2]);
		*index += 3;
		#endif
		buffer[(*index)++] = sx+ux; buffer[(*index)++] = sy+uy; buffer[(*index)++] = sz+uz; buffer[(*index)++] = texx+TEXSIZE; buffer[(*index)++] = texy+TEXSIZE;
		#ifndef DISABLELIGHTING
		calculate_light(
			x+wx, y+wy, z+wz,
			ux, uy, uz,
			-vx, -vy, -vz,
			&buffer[*index], &buffer[*index + 1], &buffer[*index + 2]);
		*index += 3;
		#endif
		buffer[(*index)++] = sx+ux+vx; buffer[(*index)++] = sy+uy+vy; buffer[(*index)++] = sz+uz+vz; buffer[(*index)++] = texx+TEXSIZE; buffer[(*index)++] = texy;
		#ifndef DISABLELIGHTING
		calculate_light(
			x+wx, y+wy, z+wz,
			ux, uy, uz,
			vx, vy, vz,
			&buffer[*index], &buffer[*index + 1], &buffer[*index + 2]);
		*index += 3;
		#endif
		buffer[(*index)++] = sx+vx; buffer[(*index)++] = sy+vy; buffer[(*index)++] = sz+vz; buffer[(*index)++] = texx; buffer[(*index)++] = texy;
		#ifndef DISABLELIGHTING
		calculate_light(
			x+wx, y+wy, z+wz,
			-ux, -uy, -uz,
			vx, vy, vz,
			&buffer[*index], &buffer[*index + 1], &buffer[*index + 2]);
		*index += 3;
		#endif
	}
	quads ++;
}

/* Add item voxel. */
static void add_tool(const unsigned short type, const int x, const int y, const int z, size_t *index) {
	// Get item texture
	unsigned int texture = frontMap[type];
	if (texture == 0) {
		return;
	}
	texture --;

	// Get coordinates of texture
	GLfloat texx = TEXSIZE*(texture%ATLASSIZE);
	GLfloat texy = TEXSIZE*(texture/ATLASSIZE);

	// Add quad
	GLfloat sy = (GLfloat)y + 0.1;

	#ifndef DISABLELIGHTING
	GLfloat red, green, blue;
	getlight(x, y, z, &red, &green, &blue);
	#endif

	buffer[(*index)++] = x; buffer[(*index)++] = sy; buffer[(*index)++] = z; buffer[(*index)++] = texx; buffer[(*index)++] = texy+TEXSIZE;
	#ifndef DISABLELIGHTING
	buffer[(*index)++] = red; buffer[(*index)++] = green; buffer[(*index)++] = blue;
	#endif
	buffer[(*index)++] = x+1; buffer[(*index)++] = sy; buffer[(*index)++] = z; buffer[(*index)++] = texx+TEXSIZE; buffer[(*index)++] = texy+TEXSIZE;
	#ifndef DISABLELIGHTING
	buffer[(*index)++] = red; buffer[(*index)++] = green; buffer[(*index)++] = blue;
	#endif
	buffer[(*index)++] = x+1; buffer[(*index)++] = sy; buffer[(*index)++] = z+1; buffer[(*index)++] = texx+TEXSIZE; buffer[(*index)++] = texy;
	#ifndef DISABLELIGHTING
	buffer[(*index)++] = red; buffer[(*index)++] = green; buffer[(*index)++] = blue;
	#endif
	buffer[(*index)++] = x; buffer[(*index)++] = sy; buffer[(*index)++] = z+1; buffer[(*index)++] = texx; buffer[(*index)++] = texy;
	#ifndef DISABLELIGHTING
	buffer[(*index)++] = red; buffer[(*index)++] = green; buffer[(*index)++] = blue;
	#endif
	quads ++;
}

/* Rebuild the level vertex buffer and load it onto the GPU. */
static void render_rebuild() {
	// Coordinates of voxel
	int x, y, z;
	// Type of voxel being appended
	unsigned short type;
	// Buffer index
	size_t index = 0;
	// Reset quad counter
	quads = 0;

	// Add voxels to buffer
	for (x = 0; x < WORLDSIZE; x ++) {
		for (y = 0; y < WORLDSIZE; y ++) {
			for (z = 0; z < WORLDSIZE; z ++) {
				// Get type of voxel
				type = world_get(x, y, z);

				// Skip hidden cross, tool voxels
				if ((specialMap[type] == RENDER_X || specialMap[type] == RENDER_TOOL) &&
						x > 0 && opaque[world_get((x-1), y, z)] &&
						x < WORLDSIZE-1 && opaque[world_get((x+1), y, z)] &&
						y > 0 && opaque[world_get(x, (y-1), z)] &&
						y < WORLDSIZE-1 && opaque[world_get(x, (y+1), z)] &&
						z > 0 && !opaque[world_get(x, y, (z-1))] &&
						z < WORLDSIZE-1 && opaque[world_get(x, y, (z+1))]) {
					continue;
				}

				if (specialMap[type] == RENDER_TOOL) {
					add_tool(type, x, y, z, &index);
				} else {
					// Top
					add_quad(type,
						x, y, z,
						1, 0, 0,
						0, 0, -1,
						0, 1, 0,
						specialMap[type],
						topMap[type],
						&index);

					// Bottom
					add_quad(type,
						x, y, z,
						-1, 0, 0,
						0, 0, -1,
						0, -1, 0,
						specialMap[type],
						bottomMap[type],
						&index);

					// Left
					add_quad(type,
						x, y, z,
						0, 0, 1,
						0, 1, 0,
						-1, 0, 0,
						specialMap[type],
						leftMap[type],
						&index);

					// Right
					add_quad(type,
						x, y, z,
						0, 0, -1,
						0, 1, 0,
						1, 0, 0,
						specialMap[type],
						rightMap[type],
						&index);

					// Front
					add_quad(type,
						x, y, z,
						1, 0, 0,
						0, 1, 0,
						0, 0, 1,
						specialMap[type],
						frontMap[type],
						&index);

					// Back
					add_quad(type,
						x, y, z,
						-1, 0, 0,
						0, 1, 0,
						0, 0, -1,
						specialMap[type],
						backMap[type],
						&index);
				}
			}
		}
	}
	glBindBuffer(GL_ARRAY_BUFFER, coords);
	glBufferData(GL_ARRAY_BUFFER, 0, NULL, GL_DYNAMIC_DRAW);
	glBufferData(GL_ARRAY_BUFFER, index*sizeof(GLfloat), &buffer[0], GL_DYNAMIC_DRAW);

	SDL_assert(glGetError() == GL_NO_ERROR);
}

static void render_sky() {
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, sky);

	glVertexPointer(3, GL_FLOAT, sizeof(GLfloat)*6, 0);
	glColorPointer(3, GL_FLOAT, sizeof(GLfloat)*6, (GLvoid*)(sizeof(GLfloat)*3));
	glDrawArrays(GL_QUADS, 0, 20);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
}

static void render_sea(const float x, const float y, const float z) {
	GLfloat red, green, blue;
	red = ((GLfloat)(insideColour[seaType]>>16)/0xff)*((GLfloat)(sunColour>>16)/0xff);
	green = ((GLfloat)((insideColour[seaType]>>8)&0xff)/0xff)*((GLfloat)((sunColour>>8)&0xff)/0xff);
	blue = ((GLfloat)(insideColour[seaType]&0xff)/0xff)*((GLfloat)(sunColour&0xff)/0xff);
	glColor3f(red, green, blue);
	glBindTexture(GL_TEXTURE_2D, seatexture);
	glBegin(GL_QUADS);
	glTexCoord2f((x-200)/16, (z-200)/16);
	glVertex3f(-200, y, -200);
	glTexCoord2f((x-200)/16, (z+200)/16);
	glVertex3f(-200, y, 200);
	glTexCoord2f((x+200)/16, (z+200)/16);
	glVertex3f(200, y, 200);
	glTexCoord2f((x+200)/16, (z-200)/16);
	glVertex3f(200, y, -200);
	glEnd();
}

void render_world() {
	if (rebuild) {
		render_rebuild();
		rebuild = 0;
	}

	#ifdef DISABLELIGHTING
	glColor3f((GLfloat)(sunColour>>16)/0xff, (GLfloat)((sunColour>>8)&0xff)/0xff, (GLfloat)(sunColour&0xff)/0xff);
	#endif
	glBindTexture(GL_TEXTURE_2D, textures);
	glBindBuffer(GL_ARRAY_BUFFER, coords);

	#ifndef DISABLELIGHTING
	glVertexPointer(3, GL_FLOAT, sizeof(GLfloat)*8, 0);
	glTexCoordPointer(2, GL_FLOAT, sizeof(GLfloat)*8, (GLvoid*)(sizeof(GLfloat)*3));
	glColorPointer(3, GL_FLOAT, sizeof(GLfloat)*8, (GLvoid*)(sizeof(GLfloat)*5));
	#else
	glVertexPointer(3, GL_FLOAT, sizeof(GLfloat)*5, 0);
	glTexCoordPointer(2, GL_FLOAT, sizeof(GLfloat)*5, (GLvoid*)(sizeof(GLfloat)*3));
	#endif
	glDrawArrays(GL_QUADS, 0, quads*4);
}

/* Draw box around the voxel the player is looking at. */
void render_selected(int x, int y, int z, int dx, int dy, int dz) {
	glColor3f(1, 1, 1);
	glBegin(GL_LINE_LOOP);
	if (dx == -1) {
		glVertex3f(x-0.1, y+0.1, z+0.1);
		glVertex3f(x-0.1, y+0.1, z+0.9);
		glVertex3f(x-0.1, y+0.9, z+0.9);
		glVertex3f(x-0.1, y+0.9, z+0.1);
	} else if (dx == 1) {
		glVertex3f(x+1.1, y+0.1, z+0.1);
		glVertex3f(x+1.1, y+0.1, z+0.9);
		glVertex3f(x+1.1, y+0.9, z+0.9);
		glVertex3f(x+1.1, y+0.9, z+0.1);
	} else if (dy == -1) {
		glVertex3f(x+0.1, y-0.1, z+0.1);
		glVertex3f(x+0.9, y-0.1, z+0.1);
		glVertex3f(x+0.9, y-0.1, z+0.9);
		glVertex3f(x+0.1, y-0.1, z+0.9);
	} else if (dy == 1) {
		glVertex3f(x+0.1, y+1.1, z+0.1);
		glVertex3f(x+0.9, y+1.1, z+0.1);
		glVertex3f(x+0.9, y+1.1, z+0.9);
		glVertex3f(x+0.1, y+1.1, z+0.9);
	} else if (dz == -1) {
		glVertex3f(x+0.1, y+0.1, z-0.1);
		glVertex3f(x+0.1, y+0.9, z-0.1);
		glVertex3f(x+0.9, y+0.9, z-0.1);
		glVertex3f(x+0.9, y+0.1, z-0.1);
	} else if (dz == 1) {
		glVertex3f(x+0.1, y+0.1, z+1.1);
		glVertex3f(x+0.9, y+0.1, z+1.1);
		glVertex3f(x+0.9, y+0.9, z+1.1);
		glVertex3f(x+0.1, y+0.9, z+1.1);
	}
	glEnd();
}

/* Draw HUD.
tool is the type of voxel being used. */
void render_ui(const unsigned short tool) {
	// Switch to 2D projection
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, width, 0, height, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Draw crosshair
	glColor3f(1, 1, 1);
	int u = 1;
	int v = 0;
	int count, z;
	for (count = 0; count < 4; count ++) {
		glBegin(GL_QUADS);
		glVertex2i(width/2 + 5*u + v, height/2 + 5*v + u);
		glVertex2i(width/2 + 20*u + v, height/2 + 20*v + u);
		glVertex2i(width/2 + 20*u - v, height/2 + 20*v - u);
		glVertex2i(width/2 + 5*u - v, height/2 + 5*v - u);
		glEnd();
		z = v;
		v = -u;
		u = z;
	}

	// Draw tool
	int tex = frontMap[tool];
	if (tex > 0) {
		tex --;
		GLfloat texx = TEXSIZE*(tex%ATLASSIZE);
		GLfloat texy = TEXSIZE*(tex/ATLASSIZE);

		glEnable(GL_TEXTURE_2D);
		glEnable(GL_ALPHA_TEST);
		glBindTexture(GL_TEXTURE_2D, textures);

		glBegin(GL_QUADS);
		glTexCoord2f(texx, texy+TEXSIZE);
		glVertex2i(width/2+8, height/2+8);
		glTexCoord2f(texx, texy);
		glVertex2i(width/2+8, height/2+40);
		glTexCoord2f(texx+TEXSIZE, texy);
		glVertex2i(width/2+40, height/2+40);
		glTexCoord2f(texx+TEXSIZE, texy+TEXSIZE);
		glVertex2i(width/2+40, height/2+8);
		glEnd();

		glDisable(GL_TEXTURE_2D);
		glDisable(GL_ALPHA_TEST);
	}

	// Draw box around tool
	glBegin(GL_LINE_LOOP);
	glVertex2i(width/2+8, height/2+8);
	glVertex2i(width/2+8, height/2+40);
	glVertex2i(width/2+40, height/2+40);
	glVertex2i(width/2+40, height/2+8);
	glEnd();

	// Change to previous matrix
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

/* Redraw with the view obscured by a solid colour.
tool is the type of voxel that is being used, which is displayed within the HUD. */
static void render_blind(const unsigned long colour, const unsigned short tool) {
	glClearColor((GLfloat)(colour>>16)/0xff, (GLfloat)((colour>>8)&0xff)/0xff, (GLfloat)(colour&0xff)/0xff, 0);
	glClear(GL_COLOR_BUFFER_BIT);
	render_ui(tool);

	SDL_GL_SwapWindow(window);

	SDL_assert(glGetError() == GL_NO_ERROR);
}

void render_normal(
		const float x, const float y, const float z,
		const float pan, const float tilt,
		const char selectionValid,
		const int sx, const int sy, const int sz,
		const char dx, const char dy, const char dz,
		const unsigned short tool) {

	// Check if camera is in a voxel. */
	int ix = floor(x);
	int iy = floor(y);
	int iz = floor(z);
	if (ix >= 0 && ix < WORLDSIZE && iy >= 0 && iy < WORLDSIZE && iz >= 0 && iz < WORLDSIZE) {
		unsigned short inside = world_get(ix, iy, iz);
		if (opaque[inside]) {
			render_blind(insideColour[inside], tool);
			return;
		}
	}

	// Check if camera is in the sea
	if (y < 0.1) {
		render_blind(insideColour[seaType], tool);
		return;
	}

	glClear(GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glRotatef(tilt, 1, 0, 0);

	render_sky();

	glRotatef(pan, 0, 1, 0);

	#ifndef DISABLEFOG
	glEnable(GL_FOG);
	#endif
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_ALPHA_TEST);
	render_sea(x, -y, z);

	glTranslatef(-x, -y, -z);

	glEnable(GL_DEPTH_TEST);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	#ifndef DISABLELIGHTING
	glEnableClientState(GL_COLOR_ARRAY);
	#endif
	render_world();
	#ifndef DISABLEFOG
	glDisable(GL_FOG);
	#endif
	#ifndef DISABLELIGHTING
	glDisableClientState(GL_COLOR_ARRAY);
	#endif
	glDisable(GL_DEPTH_TEST);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_ALPHA_TEST);

	#ifndef DISABLEHUD
	if (selectionValid) {
		render_selected(sx, sy, sz, dx, dy, dz);
	}
	render_ui(tool);
	#endif

	SDL_GL_SwapWindow(window);

	SDL_assert(glGetError() == GL_NO_ERROR);
}
