#include <SDL2/SDL_image.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "config.h"
#include "log.h"
#include "resource.h"

SDL_Window *window;
SDL_GLContext context;

char *bind_path = NULL;
char *world_path = NULL;

Mix_Chunk *sfxGrass, *sfxWood, *sfxGlass, *sfxStone, *sfxMetal, *sfxPlant, *sfxFluff, *sfxSand;

GLuint textures, seatexture;

/* Load a sound and crash on error */
static void load_sound(const char *path, Mix_Chunk **sound) {
	*sound = Mix_LoadWAV(path);
	if (*sound == NULL) {
		log_crash(Mix_GetError());
	}
}

/* Load an image as an OpenGL texture and crash on error */
static void load_texture(const char *path, GLuint *texture) {
	SDL_Surface *data = IMG_Load(path);
	if (data == NULL) {
		log_crash(IMG_GetError());
	}
	if (data->format->BytesPerPixel != 4 || data->format->Rmask != 0x000000ff) {
		log_crash("Texture is not a 32 bit RGBA image");
	}

	glGenTextures(1, texture);
	glBindTexture(GL_TEXTURE_2D, *texture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, data->w, data->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, data->pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	SDL_FreeSurface(data);
	SDL_assert(glGetError() == GL_NO_ERROR);
}

/* Set window icon and crash on error */
static void load_icon(SDL_Window *window, const char *path) {
	SDL_Surface *data = IMG_Load(path);
	if (data == NULL) {
		log_crash(IMG_GetError());
	}

	SDL_SetWindowIcon(window, data);
	SDL_FreeSurface(data);
}

/* Create save game directory and get path. */
static void get_save_path() {
	char *path = SDL_GetPrefPath("Fudge", "Build Minimal");
	size_t pathlen;

	if (path == NULL) {
		world_path = NULL;
		bind_path = NULL;

		LOG_ERROR("Unable to create save directory");
	} else {
		pathlen = strlen(path);
		world_path = malloc(pathlen + sizeof(WORLDFILE) + 1);
		bind_path = malloc(pathlen + sizeof(BINDFILE) + 1);
		if (world_path == NULL || bind_path == NULL) {
			log_crash("Failed to allocated memory");
		}

		snprintf(world_path, pathlen + sizeof(WORLDFILE) + 1, "%s%s", path, WORLDFILE);
		snprintf(bind_path, pathlen + sizeof(BINDFILE) + 1, "%s%s", path, BINDFILE);

		LOG_INFO("Save file paths:");
		LOG_INFO(world_path);
		LOG_INFO(bind_path);
	}

	SDL_free(path);
}

void resource_init() {
	if (SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO) != 0) {
		log_crash(SDL_GetError());
	}

	// SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
	// SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 4);
	window = SDL_CreateWindow(
		"Build Minimal v "VERSION,
		SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
		640, 480,
		SDL_WINDOW_OPENGL|SDL_WINDOW_RESIZABLE|SDL_WINDOW_MAXIMIZED);
	if (window == NULL) {
		log_crash(SDL_GetError());
	}

	if ((IMG_Init(IMG_INIT_PNG)&IMG_INIT_PNG) == 0) {
		log_crash(IMG_GetError());
	}

	if (Mix_Init(MIX_INIT_OGG) == -1) {
		log_crash(Mix_GetError());
	}

	if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, AUDIO_S16, 1, 4096)) {
		log_crash(Mix_GetError());
	}
	Mix_AllocateChannels(8);

	context = SDL_GL_CreateContext(window);
	if (context == NULL) {
		log_crash(SDL_GetError());
	}

	// Load OpenGL 1.5 using generated loader
	if (ogl_LoadFunctions() == ogl_LOAD_FAILED) {
		SDL_GL_DeleteContext(context);
		log_crash("Failed to initialize OpenGL 1.5\r\n");
	}

	get_save_path();

	SDL_SetRelativeMouseMode(SDL_TRUE);

	srand(time(NULL));

	load_icon(window, IMGPATH "icon.png");
	load_texture(IMGPATH "textures.png", &textures);
	load_texture(IMGPATH "sea.png", &seatexture);
	load_sound(SNDPATH "grass.ogg", &sfxGrass);
	load_sound(SNDPATH "wood.ogg", &sfxWood);
	load_sound(SNDPATH "glass.ogg", &sfxGlass);
	load_sound(SNDPATH "stone.ogg", &sfxStone);
	load_sound(SNDPATH "metal.ogg", &sfxMetal);
	load_sound(SNDPATH "plant.ogg", &sfxPlant);
	load_sound(SNDPATH "fluff.ogg", &sfxFluff);
	load_sound(SNDPATH "sand.ogg", &sfxSand);
}

void resource_quit() {
	Mix_FreeChunk(sfxGrass);
	Mix_FreeChunk(sfxWood);
	Mix_FreeChunk(sfxGlass);
	Mix_FreeChunk(sfxStone);
	Mix_FreeChunk(sfxMetal);
	Mix_FreeChunk(sfxPlant);
	Mix_FreeChunk(sfxFluff);
	Mix_FreeChunk(sfxSand);
	SDL_GL_DeleteContext(context);
	SDL_DestroyWindow(window);
	Mix_CloseAudio();
	IMG_Quit();
	free(world_path);
	free(bind_path);
	SDL_Quit();
}
