#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <SDL2/SDL.h>
#include "log.h"
#include "world.h"
#include "voxels.h"

unsigned short voxels[WORLDSIZE*WORLDSIZE*WORLDSIZE];
#ifndef DISABLELIGHTING
unsigned char light[WORLDSIZE*WORLDSIZE*WORLDSIZE*3];
/* Set when lighting needs recalculating. */
unsigned char relight = 1;

unsigned long shadow[VOXELTYPES];
unsigned long glow[VOXELTYPES];
#endif
char opaque[VOXELTYPES];
char solid[VOXELTYPES];
unsigned long insideColour[VOXELTYPES];

unsigned long skyHorizon = 0x9BCCE9;
unsigned long skyAtmosphere = 0x6AA8D3;
unsigned long skyZenith = 0x366AA3;
unsigned long sunColour = 0xFFFFFF;

unsigned short seaType = 0;

void world_init() {
	// Clear properties
	int i;
	for (i = 0; i < VOXELTYPES; i ++) {
		opaque[i] = 0;
		solid[i] = 0;
		insideColour[i] = 0x000000;
		#ifndef DISABLELIGHTING
		shadow[i] = 0;
		glow[i] = 0;
		#endif
	}
}

void world_clear() {
	int i;
	for (i = 0; i < WORLDSIZE*WORLDSIZE*WORLDSIZE; i ++) {
		voxels[i] = 0;
	}
}

char world_load(const char *path) {
	errno = 0;
	FILE *fd = fopen(path, "rb");
	if (fd == NULL) {
		LOG_PERROR("Couldn't open save file");
		return 0;
	}

	int size = fgetc(fd);
	if (size == EOF) {
		if (errno) {
			LOG_PERROR("Error while reading save file");
		} else {
			LOG_ERROR("Error while reading save file: unexpected end of file");
		}
		fclose(fd);
		return 0;
	}
	if (size != WORLDSIZE) {
		LOG_ERROR("Size of world in save file invalid");
		LOG_INFO("Has the game been compiled with a different worldsize?");
		fclose(fd);
		return 0;
	}

	// Read colours
	skyHorizon = ((unsigned long)fgetc(fd)<<16) | ((unsigned long)fgetc(fd)<<8) | (unsigned long)fgetc(fd);
	skyAtmosphere = ((unsigned long)fgetc(fd)<<16) | ((unsigned long)fgetc(fd)<<8) | (unsigned long)fgetc(fd);
	skyZenith = ((unsigned long)fgetc(fd)<<16) | ((unsigned long)fgetc(fd)<<8) | (unsigned long)fgetc(fd);
	sunColour = ((unsigned long)fgetc(fd)<<16) | ((unsigned long)fgetc(fd)<<8) | (unsigned long)fgetc(fd);

	// Read sea voxel
	seaType = ((unsigned short)fgetc(fd)<<8) | ((unsigned short)fgetc(fd));
	if (errno != 0) {
		LOG_PERROR("Error while reading save file");
		fclose(fd);
		return 0;
	}
	if (seaType >= LOADEDVOXELS) {
		LOG_ERROR("File damaged, invalid sea type");
		fclose(fd);
		return 0;
	}

	// Read voxels
	int hb, lb;
	size_t i;
	for (i = 0; i < WORLDSIZE*WORLDSIZE*WORLDSIZE; i ++) {
		hb = fgetc(fd);
		lb = fgetc(fd);
		if (hb == EOF || lb == EOF) {
			if (errno) {
				LOG_PERROR("Error while reading save file");
			} else {
				LOG_ERROR("Error while reading save file: unexpected end of file");
			}
			fclose(fd);
			return 0;
		}

		voxels[i] = (hb<<8) | lb;

		if (voxels[i] >= LOADEDVOXELS) {
			LOG_ERROR("File damaged, invalid voxel type");
			fclose(fd);
			return 0;
		}
	}

	LOG_INFO("Read save file");
	
	#ifndef DISABLELIGHTING
	relight = 1;
	#endif

	fclose(fd);
	return 1;
}

char world_save(const char *path) {
	errno = 0;
	FILE *fd = fopen(path, "wb");
	if (fd == NULL) {
		LOG_PERROR("Couldn't create save file");
		return 0;
	}

	// Write size
	if (WORLDSIZE > 255) {
		LOG_ERROR("World too large to save");
		LOG_INFO("Recompile game with worldsize < 256");
		fclose(fd);
		return 0;
	}
	if (fputc(WORLDSIZE, fd) == EOF) {
		LOG_PERROR("Error while writing save file");
		fclose(fd);
		return 0;
	}

	// Write colours
	fputc(skyHorizon>>16, fd); fputc((skyHorizon>>8)&0xff, fd); fputc(skyHorizon&0xff, fd);
	fputc(skyAtmosphere>>16, fd); fputc((skyAtmosphere>>8)&0xff, fd); fputc(skyAtmosphere&0xff, fd);
	fputc(skyZenith>>16, fd); fputc((skyZenith>>8)&0xff, fd); fputc(skyZenith&0xff, fd);
	fputc(sunColour>>16, fd); fputc((sunColour>>8)&0xff, fd); fputc(sunColour&0xff, fd);

	// Write sea voxel
	fputc(seaType>>8, fd); fputc(seaType&0xff, fd);

	if (errno != 0) {
		LOG_PERROR("Error while writing save file");
		fclose(fd);
		return 0;
	}

	size_t i;
	for (i = 0; i < WORLDSIZE*WORLDSIZE*WORLDSIZE; i ++) {
		if (fputc(voxels[i]>>8, fd) == EOF || fputc(voxels[i]&0xff, fd) == EOF) {
			LOG_PERROR("Error while writing save file");
			fclose(fd);
			return 0;
		}
	}

	LOG_INFO("Wrote save file");

	fclose(fd);
	return 1;
}

#ifndef DISABLELIGHTING
/* Calculate lighting for a voxel (recursive flood fill). */
static void world_lightvoxel(int x, int y, int z, long r, long g, long b) {
	// Apply light absorbtion
	unsigned short type = world_get(x, y, z);
	r -= (shadow[type]>>16)+AIRSHADOW;
	if (r < 0) {
		r = 0;
	}
	g -= ((shadow[type]>>8)&0xff)+AIRSHADOW;
	if (g < 0) {
		g = 0;
	}
	b -= (shadow[type]&0xff)+AIRSHADOW;
	if (b < 0) {
		b = 0;
	}

	// Apply light
	char flow = 0;
	if (light[WORLD_LIGHT(x, y, z)] < r) {
		light[WORLD_LIGHT(x, y, z)] = r;
		flow = 1;
	}
	if (light[WORLD_LIGHT(x, y, z)+1] < g) {
		light[WORLD_LIGHT(x, y, z)+1] = g;
		flow = 1;
	}
	if (light[WORLD_LIGHT(x, y, z)+2] < b) {
		light[WORLD_LIGHT(x, y, z)+2] = b;
		flow = 1;
	}
	if (!flow) {
		return;
	}

	// Recursively spread light
	if (x > 0) {
		world_lightvoxel(x-1, y, z, r, g, b);
	}
	if (x < WORLDSIZE-1) {
		world_lightvoxel(x+1, y, z, r, g, b);
	}
	if (y > 0) {
		world_lightvoxel(x, y-1, z, r, g, b);
	}
	if (y < WORLDSIZE-1) {
		world_lightvoxel(x, y+1, z, r, g, b);
	}
	if (z > 0) {
		world_lightvoxel(x, y, z-1, r, g, b);
	}
	if (z < WORLDSIZE-1) {
		world_lightvoxel(x, y, z+1, r, g, b);
	}
}
#endif

void world_relight() {
#ifndef DISABLELIGHTING
	// Ignore if lighting doesn't need to be recalculated
	if (!relight) {
		return;
	}
	relight = 0;

	// Clear light
	size_t i;
	for (i = 0; i < WORLDSIZE*WORLDSIZE*WORLDSIZE*3; i ++) {
		light[i] = 0;
	}

	// Apply sunlight

	int x, y, z;
	unsigned short type;
	// Amount of remaining sunlight
	long r, g, b;
	// Total light in voxel
	unsigned char vr, vg, vb;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (z = 0; z < WORLDSIZE; z ++) {
			r = sunColour>>16;
			g = (sunColour>>8)&0xff;
			b = sunColour&0xff;
			for (y = WORLDSIZE-1; y >= 0 ; y --) {
				type = world_get(x, y, z);
				// Apply shadow
				if (r > 0 || g > 0 || b > 0) {
					r -= shadow[type]>>16;
					if (r < 0) {
						r = 0;
					}
					g -= (shadow[type]>>8)&0xff;
					if (g < 0) {
						g = 0;
					}
					b -= shadow[type]&0xff;
					if (b < 0) {
						b = 0;
					}
				}

				// Apply glow
				vr = glow[type]>>16;
				if (r > vr) {
					vr = r;
				}
				vg = (glow[type]>>8)&0xff;
				if (g > vg) {
					vg = g;
				}
				vb = glow[type]&0xff;
				if (b > vb) {
					vb = b;
				}

				// Fill area
				if (vr > 0 || vg > 0 || vb > 0) {
					world_lightvoxel(x, y, z, vr, vg, vb);
				}
			}
		}
	}
#endif
}

void world_dirty() {
#ifndef DISABLELIGHTING
	relight = 1;
#endif
}

char world_ray(
		const float startx, const float starty, const float startz,
		const float dx, const float dy, const float dz,
		int *hitx, int *hity, int *hitz,
		char *facex, char *facey, char *facez) {

	// Set first coordinate
	(*hitx) = floor(startx);
	(*hity) = floor(starty);
	(*hitz) = floor(startz);

	// Directions to step in
	int stepX = -1;
	if (dx > 0) {
		stepX = 1;
	}
	int stepY = -1;
	if (dy > 0) {
		stepY = 1;
	}
	int stepZ = -1;
	if (dz > 0) {
		stepZ = 1;
	}

	// Time until reaching next voxel for each direction
	float tMaxX = 0;
	if (stepX > 0) {
		tMaxX = (1-startx+floor(startx))/dx;
	} else {
		tMaxX = (startx-floor(startx))/-dx;
	}
	float tMaxY = 0;
	if (stepY > 0) {
		tMaxY = (1-starty+floor(starty))/dy;
	} else {
		tMaxY = (starty-floor(starty))/-dy;
	}
	float tMaxZ = 0;
	if (stepZ > 0) {
		tMaxZ = (1-startz+floor(startz))/dz;
	} else {
		tMaxZ = (startz-floor(startz))/-dz;
	}

	// Time to move by a voxel on each axis
	float tDeltaX = fabs(1/dx);
	float tDeltaY = fabs(1/dy);
	float tDeltaZ = fabs(1/dz);

	(*facex) = 0;
	(*facey) = 0;
	(*facez) = 0;
	while (1) {
		if ((*hitx) >= 0 && (*hitx) < WORLDSIZE && (*hity) >= 0 && (*hity) < WORLDSIZE && (*hitz) >= 0 && (*hitz) < WORLDSIZE) {
			if (world_get(*hitx, *hity, *hitz)) {
				return 1;
			}
		}
		(*facex) = 0;
		(*facey) = 0;
		(*facez) = 0;
		if (tMaxX < tMaxY) {
			if (tMaxX < tMaxZ) {
				(*hitx) += stepX;
				(*facex) = -stepX;
				if (stepX > 0) {
					if ((*hitx) >= WORLDSIZE) {
						return 0;
					}
				} else {
					if ((*hitx) < 0) {
						return 0;
					}
				}
				tMaxX += tDeltaX;
			} else {
				(*hitz) += stepZ;
				(*facez) = -stepZ;
				if (stepZ > 0) {
					if ((*hitz) >= WORLDSIZE) {
						return 0;
					}
				} else {
					if ((*hitz) < 0) {
						return 0;
					}
				}
				tMaxZ += tDeltaZ;
			}
		} else {
			if (tMaxY < tMaxZ) {
				(*hity) += stepY;
				(*facey) = -stepY;
				if (stepY > 0) {
					if ((*hity) >= WORLDSIZE) {
						return 0;
					}
				} else {
					if ((*hity) < 0) {
						return 0;
					}
				}
				tMaxY += tDeltaY;
			} else {
				(*hitz) += stepZ;
				(*facez) = -stepZ;
				if (stepZ > 0) {
					if ((*hitz) >= WORLDSIZE) {
						return 0;
					}
				} else {
					if ((*hitz) < 0) {
						return 0;
					}
				}
				tMaxZ += tDeltaZ;
			}
		}
	}
	return 0;
}

